#include "vW_abwehr.h"

void Abwehr::Behave()
{

    /* Grundgeruest; 
       Wenn der Ball kickable ist wird das entsprechende Verhalten angeworfen

       Ansonsten wird zwischen eigenem Ballbesitz und gegnerischen 
       Ballbesitz unterschieden und entsprechend gehandelt.

       'onStartPos', 'gestoppt' und 'geschossen' sind temporaere Variablen 
       die hier gegebenenfalls zurueckgesetzt werden
     */

    onStartPos=false;
    

    if (Mem->BallKickable())
    {
	BallKickableBehaviour();
    }
    else 
    {
	gestoppt=false;
	geschossen=false;
	
	if (BallNotWithMyTeam())
	{
	    BallNotWithMyTeamBehaviour();
	} 
	else 
	{
	    BallWithMyTeamBehaviour();
	}
    }
}



bool Abwehr::Organizer()
{
    /* Wenn ich der Ballnaechste Abwehrspieler bin,
       bin ich der Abwehrorganisator 
       
       !!! Die Funktion wird momentan nicht benutzt, da die statische 
       Bestimmung besser laeuft !!!

       Ausserdem funktioniert Organizer() nur bei drei Abwehrspielern
    */

    bool ret = TRUE;
    
    float temp_dist = Mem->MyPos().dist(Mem->BallAbsolutePosition());
    
    for (Unum tm = 2; tm <= 4; tm++) 
    {
	if (tm == Mem->MyNumber)
	    continue;
	if (!Mem->TeammatePositionValid(tm))
	    continue;
	if (Mem->TeammateAbsolutePosition(tm).dist(Mem->BallAbsolutePosition())
	    < temp_dist)
	ret = FALSE;
    }
    return ret;
}



void Abwehr::yAusrichtung()
{
    /* 
       Hiermit bestimmt ein Abwehrspieler seine y-Ausrichtung wenn 
       genug Zeit ist um aufzuruecken.
     */



    /* Ausgangswert ist die y-Koordinate des Zentrums des Kontrolbereiches */

    float ret = ActualControlPosition.y;


    /* 
       Im Folgenden wird der Gegenspieler ermittelt der dem 
       Abwehrspieler am naechsten steht. 
    */

    Unum opp_number_temp;
    
    float opp_dist_temp = 100.0;
    float opp_dist_temp_new;
    
    
    for (Unum opp = 1; opp <= Mem->SP_team_size; opp++) 
    {
	if (!Mem->OpponentPositionValid(opp))
	    continue;

	opp_dist_temp_new = ActualControlPosition.dist(Mem->OpponentAbsolutePosition(opp));
	
	if (opp_dist_temp_new < opp_dist_temp)
	{
	    opp_dist_temp = opp_dist_temp_new;
	    opp_number_temp = opp;
	}
    }
 

    /* Wenn der naechste Spieler nicht wenigstens auf 25 an den 
       Abwehrspieler herangekommen ist orientiert, der Abwehrspieler 
       sich nicht an ihm sondern am Ball. 
       Ist der Ball weiter weg als 30 geht der Abwehrspieler ihm 
       y-maessig um die Haelfte entgegen
    
       Ist der naechste Gegner nah genug, dann nimmt der 
       Abwehrspieler dessen y-Position leicht versetzt in Richtung 
       Ball ein.

       Falls keiner der Faelle zutrifft nimmt der Spieler doch seine 
       Kontrollposition ein.
    */

    if (opp_dist_temp < 25)
    {
	ret = Mem->OpponentAbsolutePosition(opp_number_temp).y;
	if (Mem->BallAbsolutePosition().y < ret)
	    ret-=0.5;
	else
	    ret+=0.5;
    } 
    else 
	if (ActualControlPosition.x < Mem->BallAbsolutePosition().x-30)
	    ret+=(Mem->BallAbsolutePosition().y - ret)*0.5;
	
    ActualControlPosition.y = ret;
    
}


void Abwehr::ChangeControlArea()
{

    /* Wenn der Ball hinter das Zentrum des Kontrollraums gelangt ist,
       wird der Kontrollraum entsprechend nach hinten verlagert;
       Ansonsten bis auf 10 hinter dem Ball aufruecken, wenn meine maximale 
       Vorwaertsposition es zulaesst;

       Bin ich nicht der Organisator, mach ich das was der sagt */
	
    if (!organizer)
	ActualControlPosition.x = defense_line;
    else
    {
	if (Mem->BallAbsolutePosition().x<ActualControlPosition.x)
	    ActualControlPosition.x = (Mem->BallAbsolutePosition().x)-2;
	else 
	{
	    ActualControlPosition.x = Min(Mem->BallAbsolutePosition().x-10,
				       MaxForwardPos);
	}
    }

    /* Sollte die errechnete Kontroll Position nicht auf dem Feld 
       sein, wird sie so lange verschoben bis sie es wieder ist. */

    while (!(Mem->FieldRectangle.IsWithin(ActualControlPosition)))
	ActualControlPosition.x+=0.2;
    
    yAusrichtung();    
}



void Abwehr::BallKickableBehaviour()
{
    /* Kann der Ball gestoppt werden, und wurde nicht gerade
       gestoppt oder geschossen, dann erstmal stoppen */

    if(Mem->BallVelocityValid() && 
       Mem->BallAbsoluteVelocity().mod() &&
       !(gestoppt || geschossen))
    {
	stop_ball();
        gestoppt=true;
	geschossen=false;
	return;
    }

    if(goodPointToScore()){
	gestoppt =false;
	Score();
        geschossen = true;
	return;
    }

    /* Gibt es einen besser postierten Mitspieler,
       wird der angepasst */

    int TeammateToPassBall=findTeammateToPassBall();

    if(TeammateToPassBall)
    {
	pass_ball(TeammateToPassBall);
	geschossen=true;
	gestoppt = false;
	return;
    } 
    else 
    {
	/* Sind keine Gegenspieler in der Naehe erstmal Ballhalten */

	if (Mem->NumOpponentsWithin(Mem->SP_kickable_area*1.2, Mem->MyPos())==0)
	    hold_ball();
	else
	{
	    /* Ist die Schussbahn frei, soll der Ball nach vorne gekellt werden.
	       Dabei wird die naehere der beiden Eckfahnen anvisiert, 
	       um den Ball vom Zentrum fern zu halten.
	       Ist es nicht moeglich den Ball nach vorne zu spielen, soll er ins
	       Seitenaus */

	    Vector BetterCornerFlag;
	    
	    if (Mem->MyPos().dist(Mem->MarkerPosition(Mem->RM_LF_Flag))<
		Mem->MyPos().dist(Mem->MarkerPosition(Mem->RM_RF_Flag)))
		BetterCornerFlag=Mem->MarkerPosition(Mem->RM_LF_Flag);
	    else
		BetterCornerFlag=Mem->MarkerPosition(Mem->RM_RF_Flag);

	    if ((Mem->AmIInCone(50, 
			       BetterCornerFlag,
				Mem->BallAbsolutePosition())) &&
		(Mem->BallAbsolutePosition().x<
		 Mem->MarkerPosition(
		     Mem->RM_My_Goal).x+Mem->SP_penalty_area_length))
	    {
		//		kick(100,
		//     Mem->AngleToFromBody(Vector(Mem->MyPos().x,
		//			 ((Mem->MyPos().y)*2))));
		kick(100,(Mem->AngleToFromBody(Vector(Mem->MyPos().x+5,
						      ((Mem->MyPos().y)*2)))));
		geschossen=true;
		gestoppt = false;
		return;
	    }
	    else
	    {
		//	kick(100,Mem->AngleToFromBody(BetterCornerFlag));
		kick(100,(Mem->AngleToFromBody(BetterCornerFlag)));
		return;
		geschossen=true;
		gestoppt = false;
	    }
	}
    }
}



bool Abwehr::BallNotWithMyTeam()
{
    /* Wenn wir den Ball nicht haben */
    
    if (//(Mem->team_passer != Mem->MyNumber) && 
	//(Mem->team_pass_time < Mem->CurrentTime-100) && 
	!(Mem->PlayerWithBall()>0))
	return TRUE;
    else
	return FALSE;
}



void Abwehr::BallNotWithMyTeamBehaviour()
{    
    /* Wenn der Ball ueber die Mittellinie ist und kein Mitspieler naeher dran ist
       ODER
       Wenn der Ballnaechste Gegner gleichzeitig der Tornaechste Spieler ist */

    if (((Mem->NumTeammatesCloserToBall()==0) &&
	 (Mem->BallAbsolutePosition().x < 0)) ||
	(Mem->ClosestOpponentToBall() == 
	 Mem->ClosestOpponentTo(Mem->MarkerPosition(Mem->RM_My_Goal))))
	
    {
	/* Wenn mindestens ein Gegner naeher am Ball ist */
	
	if (Mem->NumOpponentsCloserToBall()>0)
	{
	    /* Wenn ich naeher am Ball bin als am Gegner dann hol ich den Ball */
	    if (Mem->MyPos().dist(Mem->BallAbsolutePosition())<
		Mem->MyPos().dist(Mem->OpponentAbsolutePosition(Mem->ClosestOpponentToBall())))
	    {
		get_ball();
		return;
	    } 

	    /* sonst versuche ich am Gegner vorbeizurennen und visiere dabei einen 
	       Punkt der zwischen Ball und Tor liegt an */
	    else
	    {
		Vector InterceptionPoint = 
		    Vector (Mem->BallAbsolutePosition().x-2,
			    Mem->BallAbsolutePosition().y - 
			    ((fabs(Mem->BallAbsolutePosition().y)/
			           Mem->BallAbsolutePosition().y)*0.5));
		

		go_to_point(InterceptionPoint, 0.1, 100, DT_none);
		face_only_neck_to_ball();
		
		return;
	    }
	} 

	/* Wenn kein Gegner naeher am Ball ist dann hol ihn dir einfach */
	else 
	{
	    get_ball();
	    return;
	}
    }
    

    /* Es sind zwar Mitspieler naeher am ballfuehrenden Spieler, aber 
       da sie vom eigenen Tor aus gesehen hinter ihm stehen, sollte
       trotzdem angegriffen werden */

    if ((Mem->BallAbsolutePosition().x < 0) &&
	(Mem->OpponentPositionValid(Mem->ClosestOpponentToBall())) &&
	(Mem->TeammatePositionValid(Mem->ClosestTeammateToBall())) &&
	(Mem->TeammateAbsolutePosition(Mem->ClosestTeammateToBall()).x>
	 Mem->OpponentAbsolutePosition(Mem->ClosestOpponentToBall()).x) && 
	(Mem->OpponentAbsolutePosition(Mem->ClosestOpponentToBall()).x>
	 Mem->MyPos().x) &&
	(Mem->NumTeammatesInCone(2,Mem->MarkerPosition(Mem->RM_My_Goal),
				 Mem->OpponentAbsolutePosition(
				     Mem->ClosestOpponentToBall()))==0)) 
    {
	get_ball();
	return;
    }

    
    /* Trifft das alles nicht zu verhalte dich als ob wir im Ballbesitz waeren */

    BallWithMyTeamBehaviour();
}



void Abwehr::BallWithMyTeamBehaviour()
{
    /* Gehe ungefhr in die Mitte deines Kontrollraumes */

    /* Wenn der Ball hinter dir ist dann renn. 
       Wenn die Mitte hinter dir liegt dann geh langsam rueckwaerts,
       ansonsten hat das Zeit */
	    
    if (Mem->MyPos().x > Mem->BallAbsolutePosition().x)
    {
	go_to_point(ActualControlPosition, 0.1, 100, DT_none);
	face_only_neck_to_ball();
	return;
    }


    float temp;
    int count = 0;
    

    /* Wenn die Abwehrspieler zu nah beieinanderstehen sollen sie wieder ein 
       wenig zurueck auf die angestammte Position */

    while ((Mem->NumTeammatesWithin(2.0, ActualControlPosition) > 0) &&
	   (count < 10))
	{
	    temp = MyStandardPosition.y - ActualControlPosition.y;
	    ActualControlPosition.y+=(fabs(temp)/temp);
	    count++;
	}
       
    /* Wenn die Mitte hinter dir liegt dann geh langsam rueckwaerts */
    
    if (Mem->MyPos().x > ActualControlPosition.x)
    {
	go_to_point(ActualControlPosition,0.1,40,DT_none);
	face_only_neck_to_ball();
	return;
    }
    
    /* ansonsten hat das Zeit */

    else
    {
	go_to_point(ActualControlPosition,0.1,20,DT_none);
	face_only_neck_to_ball();
	return;
    }
    
}


/* Manndeckungsfunktion (momentan nicht benutzbar) */


    /* Wenn sich nur ein Gegenspieler im nheren (Hlfte) Kontrollraum 
       befindet, kommt er in Manndeckung */


    /* Idee: mit turn_neck_to_relative_angle(PI), alle 50/52 o. 
       (zwei Zahlen mit einem ziemlich hohen gemeinsamen Vielfachen...)
       Zyklen um 180 Grad drehen....*/
    /* 
    if (Mem->NumOpponentsWithin(ActionRadius/2, 
				ActualControlPosition)==1)
    {
	float Faktor;
	
	if (Mem->OpponentAbsolutePosition(
	    Mem->ClosestOpponentTo(ActualControlPosition)).y>0)
	    Faktor = -0.5;
	else
	    Faktor = 0.5;
    */    
/*	GoToPointFaceOpponent(Vector(
	(Mem->OpponentAbsolutePosition(Mem->ClosestOpponentTo(ActualControlPosition)).x)-0.5,
	(Mem->OpponentAbsolutePosition(Mem->ClosestOpponentTo(ActualControlPosition)).y)-Faktor),
	Mem->ClosestOpponentTo(ActualControlPosition),50,0.1);
*/
/*	GoToPointFaceBall(Vector(
	    (Mem->OpponentAbsolutePosition(
		Mem->ClosestOpponentTo(ActualControlPosition)).x)-0.5,
	    (Mem->OpponentAbsolutePosition(
		Mem->ClosestOpponentTo(ActualControlPosition)).y)-Faktor)
			  ,50,0.1);
	
	return;
	
    }
    
    else 
*/


/*void Abwehr::HelpGoalie()
{
    Vector BetterGoalKickPoint;
    
    if (GoalKickPoint1.dist(Mem->BallAbsolutePosition())<
	GoalKickPoint2.dist(Mem->BallAbsolutePosition()))
	BetterGoalKickPoint = GoalKickPoint1;
    else 
	BetterGoalKickPoint = GoalKickPoint2;
    go_to_point(BetterGoalKickPoint, 0.1, 50, DT_none);

}*/

void Abwehr::HelpTeammate()
{

    /* Bewege dich auf deine vorher bestimmte Position, wenn es 
       zwei gibt nimm die die naeher am Ball liegt */ 

    Vector HelpPoint1=Vector(Mem->BallAbsolutePosition().x+10,
			     HelpY1);
    
    Vector HelpPoint2=Vector(Mem->BallAbsolutePosition().x+10,
			     HelpY2);

    Vector BetterHelpPoint;
    
    if (HelpPoint1.dist(Mem->BallAbsolutePosition())<
	HelpPoint2.dist(Mem->BallAbsolutePosition()))
	BetterHelpPoint = HelpPoint1;
    else 
	BetterHelpPoint =HelpPoint2;
    go_to_point(BetterHelpPoint, 0.1, 50, DT_none);
    
}



void Abwehr::CornerFormation()
{
    /* Eckballaufstellung unterschiedlich je nachdem von welcher Seite sie getreten wird */

    if (Mem->BallAbsolutePosition().y>0)
	RightCornerFormation();
    else LeftCornerFormation();
}



void Abwehr::LeftCornerFormation()
{
    
    /* Linke-Ecke Aufstellung; Je nach Spielsystem und Spielernummer wird dem Spieler 
       eine Position in der Abwehr zugewiesen */
    
    Vector RechterPfosten = Vector(Mem->MarkerPosition(Mem->RM_My_Goal).x,
				   (Mem->SP_goal_width/2));
    switch(Mem->Formation)  
    {
    case NT_VERYOFFENSIVE:
	switch(Mem->MyNumber)
	{
	case 2:
	    go_to_point(Vector(RechterPfosten.x+10,
			       (-RechterPfosten.y)-10), 0.1, 50, DT_none);
	    break;
	case 3:
	    go_to_point(RechterPfosten, 0.1, 50, DT_none);
	    break;
	}
	break;
	
    case NT_OFFENSIVE:
    case NT_MEDIUM:
	switch(Mem->MyNumber)
	{
	case 2:
	    go_to_point(Vector(RechterPfosten.x+10,
			       (-RechterPfosten.y)-10), 0.1, 50, DT_none);
	    break;
	case 3:
	    go_to_point(Vector(RechterPfosten.x+10,0), 0.1, 50, DT_none);
	    break;
	case 4:
	    go_to_point(RechterPfosten, 0.1, 50, DT_none);
	    break;
	}
	break;
	
    case NT_DEFENSIVE:
	switch(Mem->MyNumber)
	{
	case 2:
	    go_to_point(Vector(RechterPfosten.x+10,
			       -Mem->SP_penalty_area_width/2.0), 0.1, 50, DT_none);
	    break;
	case 3:
	    go_to_point(Vector(RechterPfosten.x+10,
			       -RechterPfosten.y), 0.1, 50, DT_none);
	    break;
	case 4:
	    go_to_point(Vector(RechterPfosten.x+10,
			       RechterPfosten.y), 0.1, 50, DT_none);
	    break;
	case 5:
	    go_to_point(RechterPfosten, 0.1, 50, DT_none);
	    break;
	}
	break;
	
    case NT_VERYDEFENSIVE:
	switch(Mem->MyNumber)
	{
	case 2:
	    go_to_point(Vector(RechterPfosten.x+5,
			       -(Mem->SP_pitch_width/2.0-10)), 0.1, 50, DT_none);
	    break;
	case 3:
	    go_to_point(Vector(RechterPfosten.x+10,
			       -Mem->SP_penalty_area_width/2.0), 0.1, 50, DT_none);
	    break;
	case 4:
	    go_to_point(Vector(RechterPfosten.x+10,
			       -RechterPfosten.y), 0.1, 50, DT_none);
	    break;
	case 5:
	    go_to_point(Vector(RechterPfosten.x+10,
			       RechterPfosten.y), 0.1, 50, DT_none);
	    break;
	case 6:
	    go_to_point(RechterPfosten, 0.1, 50, DT_none);
	    break;
	}
	break;
	
    case NT_CEMENTMIXER:
	switch(Mem->MyNumber)
	{
	case 2:
	    go_to_point(Vector(RechterPfosten.x+5,
			       -(Mem->SP_pitch_width/2.0-10)), 0.1, 50, DT_none);
	    break;
	case 3:
	    go_to_point(Vector(RechterPfosten.x+10,
			       -Mem->SP_penalty_area_width/2.0), 0.1, 50, DT_none);
	    break;
	case 4:
	    go_to_point(Vector(RechterPfosten.x+10,
			       -RechterPfosten.y), 0.1, 50, DT_none);
	    break;
	case 5:
	    go_to_point(Vector(RechterPfosten.x+20,0), 0.1, 50, DT_none);
	    break;
	case 6:
	    go_to_point(Vector(RechterPfosten.x+10,
			       RechterPfosten.y), 0.1, 50, DT_none);
	    break;
	case 7:
	    go_to_point(RechterPfosten, 0.1, 50, DT_none);
	    break;
	case 8:
	    go_to_point(Vector(RechterPfosten.x+10,
			       Mem->SP_penalty_area_width/2.0), 0.1, 50, DT_none);
	    break;
	}
	break;
	
    }
}




void Abwehr::RightCornerFormation()
{

    /* Rechte Ecke Formation; Analog zu links */

    Vector LinkerPfosten =Vector(Mem->MarkerPosition(Mem->RM_My_Goal).x,
				 -(Mem->SP_goal_width/2));

    switch(Mem->Formation)  
    {
    case NT_VERYOFFENSIVE:
	switch(Mem->MyNumber)
	{
	case 2:
	    go_to_point(LinkerPfosten, 0.1, 50, DT_none);
	    break;
	case 3:
	    go_to_point(Vector(LinkerPfosten.x+10,
			       (-LinkerPfosten.y)+10), 0.1, 50, DT_none);
	    break;
	}
	break;
	
    case NT_OFFENSIVE:
    case NT_MEDIUM:
	switch(Mem->MyNumber)
	{
	case 2:
	    go_to_point(LinkerPfosten, 0.1, 50, DT_none);
	    break;
	case 3:
	    go_to_point(Vector(LinkerPfosten.x+10,0), 0.1, 50, DT_none);
	    break;
	case 4:
	    go_to_point(Vector(LinkerPfosten.x+10,
			       (-LinkerPfosten.y)+10), 0.1, 50, DT_none);
	    break;
	}
	break;
	
    case NT_DEFENSIVE:
	switch(Mem->MyNumber)
	{
	case 2:
	    go_to_point(LinkerPfosten, 0.1, 50, DT_none);
	    break;
	case 3:
	    go_to_point(Vector(LinkerPfosten.x+10,
			       LinkerPfosten.y), 0.1, 50, DT_none);
	    break;
	case 4:
	    go_to_point(Vector(LinkerPfosten.x+10,
			       -LinkerPfosten.y), 0.1, 50, DT_none);
	    break;
	case 5:
	    go_to_point(Vector(LinkerPfosten.x+10,
			       Mem->SP_penalty_area_width/2.0), 0.1, 50, DT_none);
	    break;
	}
	break;
	
    case NT_VERYDEFENSIVE:
	switch(Mem->MyNumber)
	{
	case 2:
	    go_to_point(LinkerPfosten, 0.1, 50, DT_none);
	    break;
	case 3:
	    go_to_point(Vector(LinkerPfosten.x+10,
			       LinkerPfosten.y), 0.1, 50, DT_none);
	    break;
	case 4:
	    go_to_point(Vector(LinkerPfosten.x+10,
			       -LinkerPfosten.y), 0.1, 50, DT_none);
	    break;
	case 5:
	    go_to_point(Vector(LinkerPfosten.x+10,
			       Mem->SP_penalty_area_width/2.0), 0.1, 50, DT_none);
	    break;
	case 6:
	    go_to_point(Vector(LinkerPfosten.x+5,
			       Mem->SP_pitch_width/2.0-10), 0.1, 50, DT_none);
	    break;
	}
	break;
	
    case NT_CEMENTMIXER:
	switch(Mem->MyNumber)
	{
	case 2:
	    go_to_point(Vector(LinkerPfosten.x+10,
			       -Mem->SP_penalty_area_width/2.0), 0.1, 50, DT_none);
	    break;
	case 3:
	    go_to_point(LinkerPfosten, 0.1, 50, DT_none);
	    break;
	case 4:
	    go_to_point(Vector(LinkerPfosten.x+10,
			       LinkerPfosten.y), 0.1, 50, DT_none);
	    break;
	case 5:
	    go_to_point(Vector(LinkerPfosten.x+20,0), 0.1, 50, DT_none);
	    break;
	case 6:
	    go_to_point(Vector(LinkerPfosten.x+10,
			       -LinkerPfosten.y), 0.1, 50, DT_none);
	    break;
	case 7:
	    go_to_point(Vector(LinkerPfosten.x+10,
			       Mem->SP_penalty_area_width/2.0), 0.1, 50, DT_none);
	    break;
	case 8:
	    go_to_point(Vector(LinkerPfosten.x+5,
			       Mem->SP_pitch_width/2.0-10), 0.1, 50, DT_none);
	    break;
	}
	break;
	
    }/*    Vector BetterCornerPoint;
    
      if (CornerPoint1.dist(Mem->BallAbsolutePosition())<
      CornerPoint2.dist(Mem->BallAbsolutePosition()))
      BetterCornerPoint = CornerPoint1;
      else 
      BetterCornerPoint = CornerPoint2;
      go_to_point(BetterCornerPoint, 0.1, 50, DT_none);
*/
}



void Abwehr::StandardBehaviour()
{
    
    /* Standardverhalten */

    static char linientext[32];
    

    /* Wenn selbst nicht weiss wo ich bin versuch ich mich erstmal zurechtzufinden */
    if (!Mem->MyConf()) {
	Mem->LogAction2(10, "Don't know where I am, scanning field");
	scan_field_with_body();
	return;
    }


    /* Wenn ich weiss wo der Ball ist...*/
    if (Mem->BallPositionValid())
    {

	/* Wenn ich der ballnaechste Spieler bin fuehre ich 'say' aus */
	if (Mem->MyNumber == Mem->ClosestTeammateToBall())
	{
	    say("");
	} 
	

	/*Zur Sicherheit wird der Kontrollraum in jedem Zyklus zunaechst
	  auf die Ausgangsposition zurueck gesetzt*/
	
	ActualControlPosition = MyStandardPosition;
    
	/*Wenn ich Abwehrchef bin ermittele ich die Linie zum Aufruecken und rufe sie rum*/ 
	if (organizer)
	  {
	    ChangeControlArea();
	    memset(linientext,0,32);
	    sprintf(linientext, " linie %.2f", ActualControlPosition.x);
	    organize(linientext);
	  }
	/*ansonsten ermittele ich die Linie und bleib ruhig*/
	else 
	  ChangeControlArea();
  

	/* Es folgen die Verhaltensweisen der Abwehrspielern in den unterschiedlichen PlayModes*/
	switch (Mem->PlayMode){
	  
	    /* Vor Spielbeginn begeben sich die Spieler auf ihre Startpositionen 
	       und scannen das Feld */

	case PM_Before_Kick_Off:
	case PM_My_Kick_Off:
	case PM_Their_Kick_Off:
	    if (onStartPos)
		scan_field_with_body();
	    else
	    {
		move(MyStartPosition);
		onStartPos=true;
	    }    
	    break;
	    
	    
	    /*Beim eigenen Einschuss wird 'kick_in()' aufgerufen, wenn man 
	      denn der ballnaechste Spieler ist*/
	    
	case PM_My_Kick_In:
	    if (Mem->MyNumber == Mem->ClosestTeammateToBall())
	    {
		kick_in();
		break;
	    } 
	    else 
	    {
		Abwehr::Behave();
		break;
	    }
	    
	    
	    /*Bei gegnerischem Einschuss wird nahe des eigenen Tores das EckenVerhalten 
	     aufgerufen ansonsten sich normal verhalten*/
	case PM_Their_Kick_In:
	    if (Mem->BallAbsolutePosition().x>
		Mem->MarkerPosition(Mem->RM_My_Goal).x+10)
	    {
		Abwehr::Behave();
		break;
	    } 
	    else 
	    {
		CornerFormation();
		break;
	    }
	
	    
	    /*Bei gegnerischem Eckball wird das Ecken-Verhalten angewendet*/
	case PM_Their_Corner_Kick:
	    
	    CornerFormation();
	    break;



	    
	    /*Bei eigenem Torwartabschlag sollten die Spieler sich moeglichst anbieten*/
	case PM_My_Goalie_Free_Kick:     /* not a real play mode */
	    
	    HelpTeammate();
	    break;
	    

	    
	case PM_My_Free_Kick:
	    if (Mem->BallAbsolutePosition().x>-10)
	    {
		Abwehr::Behave();
		break;
	    }

	    
	case PM_My_Goal_Kick:
	case PM_My_Offside_Kick:
	    if (organizer)
	    {
		if (Mem->BallKickable())
		{
		    
		    if (Mem->CurrentTime.t<stop)
		    {
			scan_field_with_body();
			usleep(10000);
			break;
		    } 
		    else 
		    {
			BallKickableBehaviour();
			break;
		    }
		} 
		else 
		{
		    get_ball();
		    stop=Mem->CurrentTime.t+20;
		    break;
		}
		
	    }
	    else
	    {
		HelpTeammate();
		break;
	    }		
	    break;

	case PM_Their_Free_Kick:
	case PM_Their_Offside_Kick:
	case PM_Drop_Ball:
	case PM_Their_Goalie_Free_Kick:  /* not a real play mode */
	case PM_Their_Goal_Kick:
	case PM_My_Corner_Kick:
	case PM_No_Mode:
	case PM_Play_On:
	case PM_Half_Time:
	case PM_Time_Up:
	case PM_Extended_Time:
	default:
	    Abwehr::Behave();
	    break;
	}
    }
    else 
    { 
	scan_field_with_body();
	my_error("Kein Plan");
    }
}



void Baumann::Behave()
{
    StandardBehaviour();
}



void Barten::Behave()
{
    StandardBehaviour();
}


void Wicky::Behave()
{
    StandardBehaviour();
}


void Skripnik::Behave()
{
    StandardBehaviour();
}


void Tjikuzu::Behave()
{
    StandardBehaviour();
}


void Wiedener::Behave()
{
    StandardBehaviour();
}



void Banovic::Behave()
{
    StandardBehaviour();
}



void Verlaat::Behave()
{
    StandardBehaviour();
}


void Krstajic::Behave()
{
    StandardBehaviour();
}



void Thiago::Behave()
{
    StandardBehaviour();
}



void Cesar::Behave()
{
    if (Mem->PlayMode == PM_Before_Kick_Off)
	if(onStartPos)
	    scan_field_with_body();
	else
	{
	    onStartPos=true;
	    move(MyStartPosition);
	}
    else {
	if (!Mem->MyConf()) {
	    Mem->LogAction2(10, "Don't know where I am, scanning field");
	    scan_field_with_body();
	    return;
	}
	
	if (Mem->BallKickable() )
	{
	    int TeammateToPassBall=findTeammateToPassBall();
	    if(TeammateToPassBall)
	    {
		pass_ball(TeammateToPassBall);
		return;
	    }
	    // kick(100,Mem->MarkerAngleFromBody(Mem->RM_Their_Goal));
	    kick(100,(Mem->MarkerAngleFromBody(Mem->RM_Their_Goal)));
	}
	else 
	{
	    if(Mem->BallInOwnPenaltyArea())
		get_ball();
	    else
		GoToPointFaceBall(MyStandardPosition, 50, 2.5);
	}
    }
}

     




